@@ -1,14 +1,18 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+from __future__ import division
+
 import datetime
 
 from sql import Null
+from sql.conditionals import Case
 
-from trytond.model import ModelView, ModelSQL, fields
+from trytond.model import ModelView, ModelSQL, fields, Unique
 from trytond.pyson import Eval
 from trytond import backend
 from trytond.transaction import Transaction
 from trytond.pool import Pool
+from trytond.tools import reduce_ids, grouped_slice
 
 __all__ = ['Work']
 
@@ -16,18 +20,19 @@ __all__ = ['Work']
 class Work(ModelSQL, ModelView):
     'Work Effort'
     __name__ = 'project.work'
-    _rec_name = 'work'
-    work = fields.Many2One('timesheet.work', 'Work', required=True,
-            ondelete='CASCADE')
-    active = fields.Function(fields.Boolean('Active'),
-        'get_active', setter='set_active', searcher='search_active')
+    name = fields.Char('Name', required=True, select=True)
+    work = fields.Many2One('timesheet.work', 'Work', ondelete='CASCADE',
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ],
+        depends=['company'])
     type = fields.Selection([
             ('project', 'Project'),
             ('task', 'Task')
             ],
         'Type', required=True, select=True)
-    company = fields.Function(fields.Many2One('company.company', 'Company'),
-        'on_change_with_company', searcher='search_comany')
+    company = fields.Many2One('company.company', 'Company', required=True,
+        select=True)
     party = fields.Many2One('party.party', 'Party',
         states={
             'invisible': Eval('type') != 'project',
@@ -41,8 +46,9 @@ class Work(ModelSQL, ModelView):
         fields.Boolean('Available on timesheets'),
         'on_change_with_timesheet_available')
     timesheet_duration = fields.Function(fields.TimeDelta('Duration',
-            'company_work_time', help="Total time spent on this work"),
-        'on_change_with_timesheet_duration')
+            'company_work_time',
+            help="Total time spent on this work and the sub-works"),
+        'get_total')
     effort_duration = fields.TimeDelta('Effort', 'company_work_time',
         states={
             'invisible': Eval('type') != 'task',
@@ -50,11 +56,40 @@ class Work(ModelSQL, ModelView):
     total_effort = fields.Function(fields.TimeDelta('Total Effort',
             'company_work_time',
             help="Estimated total effort for this work and the sub-works"),
-        'get_total_effort')
+        'get_total')
+    progress = fields.Float('Progress',
+        domain=['OR',
+            ('progress', '=', None),
+            [
+                ('progress', '>=', 0),
+                ('progress', '<=', 1),
+                ],
+            ],
+        states={
+            'invisible': Eval('type') != 'task',
+            },
+        depends=['type'],
+        help='Estimated progress for this work')
+    total_progress = fields.Function(fields.Float('Total Progress',
+            help='Estimated total progress for this work and the sub-works',
+            states={
+                'invisible': Eval('total_progress', None) == None,
+                }),
+        'get_total')
     comment = fields.Text('Comment')
-    parent = fields.Function(fields.Many2One('project.work', 'Parent'),
-            'get_parent', setter='set_parent', searcher='search_parent')
-    children = fields.One2Many('project.work', 'parent', 'Children')
+    parent = fields.Many2One('project.work', 'Parent',
+        left='left', right='right', ondelete='RESTRICT',
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ],
+        depends=['company'])
+    left = fields.Integer('Left', required=True, select=True)
+    right = fields.Integer('Right', required=True, select=True)
+    children = fields.One2Many('project.work', 'parent', 'Children',
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ],
+        depends=['company'])
     state = fields.Selection([
             ('opened', 'Opened'),
             ('done', 'Done'),
@@ -64,19 +99,27 @@ class Work(ModelSQL, ModelView):
     @staticmethod
     def order_sequence(tables):
         table, _ = tables[None]
-        return [table.sequence == Null, table.sequence]
+        return [Case((table.sequence == Null, 0), else_=1), table.sequence]
 
     @staticmethod
     def default_type():
         return 'task'
 
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
     @staticmethod
     def default_state():
         return 'opened'
 
-    @staticmethod
-    def default_effort():
-        return 0.0
+    @classmethod
+    def default_left(cls):
+        return 0
+
+    @classmethod
+    def default_right(cls):
+        return 0
 
     @classmethod
     def __register__(cls, module_name):
@@ -90,6 +133,9 @@ class Work(ModelSQL, ModelView):
 
         migrate_sequence = (not table_project_work.column_exist('sequence')
             and table_timesheet_work.column_exist('sequence'))
+        add_parent = not table_project_work.column_exist('parent')
+        add_company = not table_project_work.column_exist('company')
+        add_name = not table_project_work.column_exist('name')
 
         super(Work, cls).__register__(module_name)
 
@@ -119,11 +165,49 @@ class Work(ModelSQL, ModelView):
                         where=project.id == id_))
             table_project_work.drop_column('effort')
 
+        # Migration from 3.6: add parent, company, drop required on work,
+        # fill name
+        if add_parent:
+            second_project = cls.__table__()
+            query = project.join(timesheet,
+                condition=project.work == timesheet.id
+                ).join(second_project,
+                    condition=timesheet.parent == second_project.work
+                    ).select(project.id, second_project.id)
+            cursor.execute(*query)
+            for id_, parent in cursor.fetchall():
+                cursor.execute(*project.update(
+                        [project.parent],
+                        [parent],
+                        where=project.id == id_))
+            cls._rebuild_tree('parent', None, 0)
+        if add_company:
+            cursor.execute(*project.join(timesheet,
+                    condition=project.work == timesheet.id
+                    ).select(project.id, timesheet.company))
+            for id_, company in cursor.fetchall():
+                cursor.execute(*project.update(
+                        [project.company],
+                        [company],
+                        where=project.id == id_))
+        table_project_work.not_null_action('work', action='remove')
+        if add_name:
+            cursor.execute(*project.join(timesheet,
+                    condition=project.work == timesheet.id
+                    ).select(project.id, timesheet.name))
+            for id_, name in cursor.fetchall():
+                cursor.execute(*project.update(
+                        [project.name],
+                        [name],
+                        where=project.id == id_))
+
     @classmethod
     def __setup__(cls):
         super(Work, cls).__setup__()
+        t = cls.__table__()
         cls._sql_constraints += [
-            ('work_uniq', 'UNIQUE(work)', 'There should be only one '
+            ('work_uniq', Unique(t, t.work),
+                'There should be only one '
                 'timesheet work by task/project.'),
             ]
         cls._order.insert(0, ('sequence', 'ASC'))
@@ -156,28 +240,44 @@ class Work(ModelSQL, ModelView):
                             'child': child.rec_name,
                             })
 
+    @property
+    def effort_hours(self):
+        if not self.effort_duration:
+            return 0
+        return self.effort_duration.total_seconds() / 60 / 60
+
+    @property
+    def total_effort_hours(self):
+        if not self.total_effort:
+            return 0
+        return self.total_effort.total_seconds() / 60 / 60
+
+    @property
+    def timesheet_duration_hours(self):
+        if not self.timesheet_duration:
+            return 0
+        return self.timesheet_duration.total_seconds() / 60 / 60
+
     def get_rec_name(self, name):
-        return self.work.name
-
-    @staticmethod
-    def default_active():
-        return True
-
-    def get_active(self, name):
-        return self.work.active
+        if self.parent:
+            return self.parent.get_rec_name(name) + '\\' + self.name
+        else:
+            return self.name
 
     @classmethod
-    def set_active(cls, works, name, value):
-        pool = Pool()
-        Work = pool.get('timesheet.work')
-
-        Work.write([p.work for p in works], {
-                'active': value,
-                })
-
-    @classmethod
-    def search_active(cls, name, clause):
-        return [('work.active',) + tuple(clause[1:])]
+    def search_rec_name(cls, name, clause):
+        if isinstance(clause[2], basestring):
+            values = clause[2].split('\\')
+            values.reverse()
+            domain = []
+            field = 'name'
+            for name in values:
+                domain.append((field, clause[1], name.strip()))
+                field = 'parent.' + field
+        else:
+            domain = [('name',) + tuple(clause[1:])]
+        ids = [w.id for w in cls.search(domain, order=[])]
+        return [('parent', 'child_of', ids)]
 
     @fields.depends('work')
     def on_change_with_company(self, name=None):
@@ -191,114 +291,9 @@ class Work(ModelSQL, ModelView):
     def on_change_with_timesheet_available(self, name=None):
         return self.work.timesheet_available if self.work else None
 
-    @fields.depends('work')
-    def on_change_with_timesheet_duration(self, name=None):
-        return self.work.duration if self.work else None
-
     @classmethod
-    def get_parent(cls, project_works, name):
-        parents = dict.fromkeys([w.id for w in project_works], None)
-
-        # ptw2pw is "parent timesheet work to project works":
-        ptw2pw = {}
-        for project_work in project_works:
-            if not project_work.work.parent:
-                continue
-            if project_work.work.parent.id in ptw2pw:
-                ptw2pw[project_work.work.parent.id].append(project_work.id)
-            else:
-                ptw2pw[project_work.work.parent.id] = [project_work.id]
-
-        with Transaction().set_context(active_test=False):
-            parent_projects = cls.search([
-                    ('work', 'in', ptw2pw.keys()),
-                    ])
-        for parent_project in parent_projects:
-            if parent_project.work.id in ptw2pw:
-                child_projects = ptw2pw[parent_project.work.id]
-                for child_project in child_projects:
-                    parents[child_project] = parent_project.id
-
-        return parents
-
-    @classmethod
-    def set_parent(cls, project_works, name, value):
-        TimesheetWork = Pool().get('timesheet.work')
-        if value:
-            project_works.append(cls(value))
-            child_timesheet_works = [x.work for x in project_works[:-1]]
-            parent_timesheet_work_id = project_works[-1].work.id
-        else:
-            child_timesheet_works = [x.work for x in project_works]
-            parent_timesheet_work_id = None
-
-        TimesheetWork.write(child_timesheet_works, {
-                'parent': parent_timesheet_work_id
-                })
-
-    @classmethod
-    def search_parent(cls, name, domain):
-        TimesheetWork = Pool().get('timesheet.work')
-
-        project_work_domain = []
-        timesheet_work_domain = []
-        if domain[0].startswith('parent.'):
-            project_work_domain.append(
-                    (domain[0].replace('parent.', ''),)
-                    + domain[1:])
-        elif domain[0] == 'parent':
-            timesheet_work_domain.append(domain)
-
-        # ids timesheet_work_domain in operand are project_work ids,
-        # we need to convert them to timesheet_work ids
-        operands = set()
-        for _, _, operand in timesheet_work_domain:
-            if (isinstance(operand, (int, long))
-                    and not isinstance(operand, bool)):
-                operands.add(operand)
-            elif isinstance(operand, list):
-                for o in operand:
-                    if isinstance(o, (int, long)) and not isinstance(o, bool):
-                        operands.add(o)
-        pw2tw = {}
-        if operands:
-            operands = list(operands)
-            # filter out non-existing ids:
-            operands = cls.search([
-                    ('id', 'in', operands)
-                    ])
-            # create project_work > timesheet_work mapping
-            for pw in operands:
-                pw2tw[pw.id] = pw.work.id
-
-            for i, d in enumerate(timesheet_work_domain):
-                if isinstance(d[2], (int, long)):
-                    new_d2 = pw2tw.get(d[2], 0)
-                elif isinstance(d[2], list):
-                    new_d2 = []
-                    for item in d[2]:
-                        item = pw2tw.get(item, 0)
-                        new_d2.append(item)
-                timesheet_work_domain[i] = (d[0], d[1], new_d2)
-
-        if project_work_domain:
-            project_works = cls.search(project_work_domain)
-            timesheet_work_domain.append(
-                ('id', 'in', [pw.work.id for pw in project_works]))
-
-        tw_ids = [tw.id for tw in TimesheetWork.search(timesheet_work_domain)]
-
-        return [('work', 'in', tw_ids)]
-
-    @classmethod
-    def sum_tree(cls, works, getter):
-        result = {}
-        parents = {}
-        for work in works:
-            result[work.id] = getter(work)
-            parent = work.parent
-            if parent:
-                parents[work.id] = parent.id
+    def sum_tree(cls, works, values, parents):
+        result = values.copy()
         works = set((w.id for w in works))
         leafs = works - set(parents.itervalues())
         while leafs:
@@ -318,13 +313,54 @@ class Work(ModelSQL, ModelView):
         return result
 
     @classmethod
-    def get_total_effort(cls, works, name):
+    def get_total(cls, works, names):
+        cursor = Transaction().cursor
+        table = cls.__table__()
+
         works = cls.search([
                 ('parent', 'child_of', [w.id for w in works]),
-                ('active', '=', True),
-                ]) + works
-        return cls.sum_tree(works,
-            lambda w: w.effort_duration or datetime.timedelta())
+                ])
+        work_ids = [w.id for w in works]
+        parents = {}
+        for sub_ids in grouped_slice(work_ids):
+            where = reduce_ids(table.id, sub_ids)
+            cursor.execute(*table.select(table.id, table.parent,
+                    where=where))
+            parents.update(cursor.fetchall())
+
+        if 'total_progress' in names and 'total_effort' not in names:
+            names = list(names)
+            names.append('total_effort')
+
+        result = {}
+        for name in names:
+            values = getattr(cls, '_get_%s' % name)(works)
+            result[name] = cls.sum_tree(works, values, parents)
+
+        if 'total_progress' in names:
+            total_progress = result['total_progress']
+            total_effort = result['total_effort']
+            for work in works:
+                if total_effort[work.id]:
+                    total_progress[work.id] /= (
+                        total_effort[work.id].total_seconds() / 60 / 60)
+                else:
+                    total_effort[work.id] = None
+        return result
+
+    @classmethod
+    def _get_total_effort(cls, works):
+        return {w.id: w.effort_duration or datetime.timedelta() for w in works}
+
+    @classmethod
+    def _get_timesheet_duration(cls, works):
+        return {w.id: (w.work.duration if w.work and w.work.duration
+                else datetime.timedelta())
+            for w in works}
+
+    @classmethod
+    def _get_total_progress(cls, works):
+        return {w.id: w.effort_hours * (w.progress or 0) for w in works}
 
     @classmethod
     def copy(cls, project_works, default=None):
@@ -340,11 +376,12 @@ class Work(ModelSQL, ModelView):
         timesheet_default['children'] = None
         new_project_works = []
         for project_work in project_works:
-            timesheet_work, = TimesheetWork.copy([project_work.work],
-                default=timesheet_default)
             pwdefault = default.copy()
             pwdefault['children'] = None
-            pwdefault['work'] = timesheet_work.id
+            if project_work.work:
+                timesheet_work, = TimesheetWork.copy([project_work.work],
+                    default=timesheet_default)
+                pwdefault['work'] = timesheet_work.id
             new_project_works.extend(super(Work, cls).copy([project_work],
                     default=pwdefault))
         return new_project_works
@@ -354,11 +391,12 @@ class Work(ModelSQL, ModelView):
         TimesheetWork = Pool().get('timesheet.work')
 
         # Get the timesheet works linked to the project works
-        timesheet_works = [pw.work for pw in project_works]
+        timesheet_works = [pw.work for pw in project_works if pw.work]
 
         super(Work, cls).delete(project_works)
 
-        TimesheetWork.delete(timesheet_works)
+        if timesheet_works:
+            TimesheetWork.delete(timesheet_works)
 
     @classmethod
     def search_global(cls, text):
